@@ -1,16 +1,15 @@
-from typing import TypedDict
+#from typing import TypedDict
+import copy
 from typing_extensions import Unpack
 
 from warp_beacon.jobs.upload_job import UploadJob
 from warp_beacon.jobs.abstract import AbstractJob, JobSettings
 
-import logging
+#import logging
 
 class DownloadJob(AbstractJob):
 	def __init__(self, **kwargs: Unpack[JobSettings]) -> None:
 		super(DownloadJob, self).__init__(**kwargs)
-	def __del__(self) -> None:
-		pass
 
 	@staticmethod
 	def build(**kwargs: Unpack[JobSettings]) -> "DownloadJob":
@@ -18,9 +17,9 @@ class DownloadJob(AbstractJob):
 	
 	def to_upload_job(self, **kwargs: Unpack[JobSettings]) -> AbstractJob:
 		d = self.to_dict()
-		d.update(kwargs)
+		d.update(copy.deepcopy(kwargs))
 		if "media_collection" in d:
-			for index, chunk in enumerate(d["media_collection"]):
-				for k, v in enumerate(d["media_collection"][index]):
+			for index, _ in enumerate(d["media_collection"]):
+				for k, _ in enumerate(d["media_collection"][index]):
 					d["media_collection"][index][k] = UploadJob.build(**d["media_collection"][index][k])
 		return UploadJob.build(**d)
\ No newline at end of file
