@@ -1,17 +1,17 @@
 import os
 import signal
+import asyncio
+from typing import Optional, Union
 
 import logging
-
 import html
-
 import uvloop
 
 from pyrogram import Client, filters
 from pyrogram.enums import ParseMode, ChatType
 from pyrogram.handlers import MessageHandler, CallbackQueryHandler
-from pyrogram.types import Message, InputMedia, InputMediaAudio, InputMediaPhoto, InputMediaVideo, InputMediaAnimation, InputMediaDocument, InlineKeyboardButton, InlineKeyboardMarkup
-from pyrogram.errors import RPCError, FloodWait, NetworkMigrate, BadRequest, MultiMediaTooLong, MessageIdInvalid
+from pyrogram.types import InputMediaAudio, InputMediaPhoto, InputMediaVideo, InputMediaAnimation, InlineKeyboardButton, InlineKeyboardMarkup
+from pyrogram.errors import NetworkMigrate, BadRequest, MultiMediaTooLong, MessageIdInvalid, FloodWait
 
 import warp_beacon
 from warp_beacon.__version__ import __version__
@@ -26,8 +26,11 @@ from warp_beacon.jobs import Origin
 from warp_beacon.telegram.utils import Utils
 from warp_beacon.telegram.caption_shortener import CaptionShortner
 from warp_beacon.scheduler.scheduler import IGScheduler
+from warp_beacon.telegram.edit_message import EditMessage
+from warp_beacon.telegram.download_status import DownloadStatus
 
 class Bot(object):
+	should_exit = None
 	storage = None
 	uploader = None
 	downloader = None
@@ -36,20 +39,21 @@ class Bot(object):
 	handlers = None
 	placeholder = None
 	scheduler = None
+	me = None
+	edit_message = None
+	download_status = None
 
 	def __init__(self, tg_bot_name: str, tg_token: str, tg_api_id: str, tg_api_hash: str) -> None:
 		# Enable logging
 		logging.basicConfig(
-			format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
+			format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
+			level=logging.INFO
 		)
 
 		logging.getLogger("pyrogram").setLevel(logging.ERROR)
-
-		db_connect = DBClient()
-		self.storage = Storage(db_connect)
-
 		logging.info("Starting Warp Beacon version '%s' ...", __version__)
-
+		self.storage = Storage(DBClient())
+		self.should_exit = asyncio.Event()
 		workers_amount = min(32, os.cpu_count() + 4)
 
 		uvloop.install()
@@ -60,36 +64,30 @@ class Bot(object):
 			api_id=tg_api_id,
 			api_hash=tg_api_hash,
 			workdir='/var/warp_beacon',
-			workers=int(os.environ.get("TG_WORKERS_POOL_SIZE", default=workers_amount))
+			workers=int(os.environ.get("TG_WORKERS_POOL_SIZE", default=workers_amount)),
+			ipv6=os.environ.get("TG_IPV6", default="false").lower() in ("1", "true", "yes")
 		)
 
-		this = self
-		def __terminator() -> None:
-			this.stop()
-
-		stop_signals = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT)
-		for sig in stop_signals:
-			self.client.loop.add_signal_handler(sig, __terminator)
+		self.editor = EditMessage(self.client)
+		self.handlers = Handlers(self)
 
 		self.uploader = AsyncUploader(
 			storage=self.storage,
+			upload_wrapper=self.handlers.upload_wrapper,
 			admin_message_callback=self.send_text_to_admin,
+			request_yt_auth_callback=self.request_yt_auth,
 			pool_size=int(os.environ.get("UPLOAD_POOL_SIZE", default=workers_amount)),
 			loop=self.client.loop
 		)
+		self.download_status = DownloadStatus(self.client)
 		self.downloader = warp_beacon.scraper.AsyncDownloader(
 			workers_count=int(os.environ.get("WORKERS_POOL_SIZE", default=workers_amount)),
-			uploader=self.uploader
+			uploader=self.uploader,
+			pipe_connection=self.download_status.child_conn
 		)
 
 		self.scheduler = IGScheduler(self.downloader)
 
-		self.downloader.start()
-		self.uploader.start()
-		self.scheduler.start()
-
-		self.handlers = Handlers(self)
-
 		self.client.add_handler(MessageHandler(self.handlers.start, filters.command("start")))
 		self.client.add_handler(MessageHandler(self.handlers.help, filters.command("help")))
 		self.client.add_handler(MessageHandler(self.handlers.random, filters.command("random")))
@@ -100,20 +98,38 @@ class Bot(object):
 
 		self.placeholder = PlaceholderMessage(self)
 
-		self.client.run()
-
 	def __del__(self) -> None:
 		self.stop()
-		logging.info("Warp Beacon terminated.")
 
 	def start(self) -> None:
-		self.client.run()
+		self.client.run(self.run())
+
+	async def run(self) -> None:
+		loop = asyncio.get_running_loop()
+		for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGABRT):
+			loop.add_signal_handler(sig, self.should_exit.set)
+		async with self.client:
+			self.me = await self.client.get_me()
+			self.client.me = self.me
+			if self.me.is_premium:
+				os.environ["TG_PREMIUM"] = "true"
+			self.downloader.start()
+			self.uploader.start()
+			loop.add_reader(self.download_status.status_pipe.fileno(), self.download_status.on_status)
+			self.scheduler.start()
+			logging.info("Warp Beacon version '%s' started", __version__)
+			await self.should_exit.wait()
+
+		self.stop()
+		logging.info("Warp Beacon version '%s' terminated.", __version__)
 
 	def stop(self) -> None:
 		logging.info("Warp Beacon is terminating. This may take a while ...")
 		self.scheduler.stop()
 		self.downloader.stop_all()
 		self.uploader.stop_all()
+		if self.client and self.client.is_initialized and self.client.is_connected:
+			asyncio.run_coroutine_threadsafe(self.client.stop(), self.client.loop)
 
 	async def send_text(self, chat_id: int, text: str, reply_id: int = None) -> int:
 		try:
@@ -130,7 +146,31 @@ class Bot(object):
 
 		return 0
 
-	async def send_text_to_admin(self, text: str, account_admins: str = None, yt_auth: bool = False) -> list[int]:
+	async def request_yt_auth(self) -> None:
+		from warp_beacon.yt_auth import YtAuth
+		acc_index, acc = self.downloader.acc_selector.get_current_for_module(Origin.YOUTUBE)
+		yt_auth = YtAuth(account_index=acc_index)
+		data = yt_auth.fetch_token()
+		if all(data.values()):
+			if yt_auth.store_device_code(data["device_code"]):
+				reply_markup = InlineKeyboardMarkup(
+					[
+						[
+							InlineKeyboardButton("✅ Done", callback_data=f"confirm_yt_auth:{acc_index}")
+						]
+					]
+				)
+				await self.send_text_to_admin(
+					f"Please open {data['verification_url']} and input code <code>{data['user_code']}</code>.\n\n"
+					"Please authorize with a Google account with verified age.\n"
+					"This will allow you to avoid error the <b>AgeRestrictedError</b> when accessing some content.",
+					account_admins=acc.get("account_admins", None),
+					reply_markup=reply_markup
+				)
+		else:
+			logging.error("Wrong YT auth dataset: '%s'", str(data))
+
+	async def send_text_to_admin(self, text: str, account_admins: str = None, reply_markup: Optional[Union[InlineKeyboardMarkup]] = None) -> list[int]:
 		try:
 			admins = None
 			if account_admins:
@@ -145,14 +185,8 @@ class Bot(object):
 			for adm in admins_array:
 				adm = adm.strip()
 				msg_opts = {"chat_id": adm, "text": text, "parse_mode": ParseMode.HTML}
-				if yt_auth:
-					msg_opts["reply_markup"] = InlineKeyboardMarkup(
-						[
-							[
-								InlineKeyboardButton("✅ Done", callback_data="auth_process_done")
-							]
-						]
-					)
+				if reply_markup:
+					msg_opts["reply_markup"] = reply_markup
 				message_reply = await self.client.send_message(**msg_opts)
 				msg_ids.append(message_reply.id)
 			return msg_ids
@@ -162,23 +196,24 @@ class Bot(object):
 
 		return []
 
-	def build_signature_caption(self, job: UploadJob) -> str:
+	def build_signature_caption(self, job: UploadJob, override_canonical_name: str = None) -> str:
 		caption = ""
 		is_group = job.chat_type in (ChatType.GROUP, ChatType.SUPERGROUP)
 		is_youtube = job.job_origin in (Origin.YOUTUBE, Origin.YT_SHORTS)
-		if job.canonical_name:
-			if is_group and not is_youtube and CaptionShortner.need_short(job.canonical_name):
-				caption = f"{html.escape(CaptionShortner.smart_truncate_html(job.canonical_name))} ..."
+		effective_canonical_name = override_canonical_name or job.canonical_name
+		if effective_canonical_name:
+			if is_group and not is_youtube and CaptionShortner.need_short(effective_canonical_name):
+				caption = f"{html.escape(CaptionShortner.smart_truncate_html(effective_canonical_name))} ..."
 				job.short_text = True
 			elif is_group:
 				# short enough
-				caption = html.escape(job.canonical_name)
+				caption = html.escape(effective_canonical_name)
 			# Captions only for YouTube in groups; empty otherwise.
 			# Might be too long for Telegram, so we skip them.
 			if is_youtube and is_group:
-				caption = f"<b>{html.escape(job.canonical_name)}</b>"
+				caption = f"<b>{html.escape(effective_canonical_name)}</b>"
 		if is_group:
-			if job.canonical_name:
+			if effective_canonical_name:
 				caption += "\n—\n"
 			if job.message_leftover:
 				caption += job.message_leftover
@@ -187,11 +222,11 @@ class Bot(object):
 				if job.source_username.startswith("<a href="):
 					caption += f"Requested by <b>{job.source_username}</b>"
 				else:
-					caption += f"Requested by <b>@{job.source_username}</b>"
+					caption += f"Requested by <b>@{html.escape(job.source_username)}</b>"
 			if job.source_username and job.url:
 				caption += " | "
 			if job.url:
-				caption += f'<a href="{job.url}">source link</a>'
+				caption += f'<a href="{html.escape(job.url)}">source link</a>'
 
 		return caption
 
@@ -203,11 +238,11 @@ class Bot(object):
 					args["media"] = InputMediaVideo(
 						media=job.tg_file_id.replace(":video", ''),
 						supports_streaming=True,
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["video"] = job.tg_file_id.replace(":video", '')
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
 			else:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaVideo(
@@ -217,7 +252,7 @@ class Bot(object):
 						height=job.media_info["height"],
 						duration=round(job.media_info["duration"]),
 						thumb=job.media_info["thumb"],
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["video"] = job.local_media_path
@@ -226,34 +261,36 @@ class Bot(object):
 					args["height"] = job.media_info["height"]
 					args["duration"] = round(job.media_info["duration"])
 					args["thumb"] = job.media_info["thumb"]
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
 
-				args["file_name"] = "downloaded_via_warp_beacon_bot%s" % (os.path.splitext(job.local_media_path)[-1])
+				args["file_name"] = os.path.basename(job.local_media_path)
 		elif job.media_type == JobType.IMAGE:
 			if job.tg_file_id:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaPhoto(
 						media=job.tg_file_id.replace(":image", ''),
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["photo"] = job.tg_file_id.replace(":image", '')
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
 			else:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaPhoto(
 						media=job.local_media_path,
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["photo"] = job.local_media_path
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
+
+				args["file_name"] = os.path.basename(job.local_media_path)
 		elif job.media_type == JobType.AUDIO:
 			if job.tg_file_id:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaAudio(
 						media=job.tg_file_id.replace(":audio", ''),
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["audio"] = job.tg_file_id.replace(":audio", '')
@@ -265,7 +302,7 @@ class Bot(object):
 						thumb=job.media_info["thumb"],
 						duration=round(job.media_info["duration"]),
 						title=job.canonical_name,
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["audio"] = job.local_media_path
@@ -273,17 +310,19 @@ class Bot(object):
 					args["thumb"] = job.media_info["thumb"]
 					args["duration"] = round(job.media_info["duration"])
 					args["title"] = job.canonical_name
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
+
+				args["file_name"] = os.path.basename(job.local_media_path)
 		elif job.media_type == JobType.ANIMATION:
 			if job.tg_file_id:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaAnimation(
 						media=job.tg_file_id.replace(":animation", ''),
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["animation"] = job.tg_file_id.replace(":animation", '')
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
 			else:
 				if job.placeholder_message_id:
 					args["media"] = InputMediaAnimation(
@@ -292,7 +331,7 @@ class Bot(object):
 						duration=round(job.media_info["duration"]),
 						width=job.media_info["width"],
 						height=job.media_info["height"],
-						caption=self.build_signature_caption(job)
+						caption=self.build_signature_caption(job, job.message_text)
 					)
 				else:
 					args["animation"] = job.local_media_path
@@ -300,7 +339,9 @@ class Bot(object):
 					args["height"] = job.media_info["height"]
 					args["duration"] = round(job.media_info["duration"])
 					args["thumb"] = job.media_info["thumb"]
-					args["caption"] = self.build_signature_caption(job)
+					args["caption"] = self.build_signature_caption(job, job.message_text)
+
+				args["file_name"] = os.path.basename(job.local_media_path)
 		elif job.media_type == JobType.COLLECTION:
 			if job.tg_file_id:
 				args["media"] = []
@@ -311,11 +352,11 @@ class Bot(object):
 						ctype = JobType[mtype.upper()]
 						ptr = None
 						if ctype == JobType.VIDEO:
-							ptr = InputMediaVideo(media=tg_id, caption=self.build_signature_caption(job))
+							ptr = InputMediaVideo(media=tg_id, caption=self.build_signature_caption(job, job.message_text))
 						elif ctype == JobType.IMAGE:
-							ptr = InputMediaPhoto(media=tg_id, caption=self.build_signature_caption(job))
+							ptr = InputMediaPhoto(media=tg_id, caption=self.build_signature_caption(job, job.message_text))
 						elif ctype == JobType.ANIMATION:
-							ptr = InputMediaAnimation(media=tg_id, caption=self.build_signature_caption(job))
+							ptr = InputMediaAnimation(media=tg_id, caption=self.build_signature_caption(job, job.message_text))
 						tg_chunk.append(ptr)
 
 					args["media"].append(tg_chunk)
@@ -332,17 +373,30 @@ class Bot(object):
 								height=j.media_info["height"],
 								duration=round(j.media_info["duration"]),
 								thumb=j.media_info["thumb"],
-								caption=self.build_signature_caption(job)
+								caption=self.build_signature_caption(job, j.canonical_name)
 							)
 							tg_chunk.append(vid)
 						elif j.media_type == JobType.IMAGE:
 							photo = InputMediaPhoto(
 								media=j.local_media_path,
-								caption=self.build_signature_caption(job)
+								caption=self.build_signature_caption(job, j.canonical_name)
 							)
 							tg_chunk.append(photo)
+						elif j.media_type == JobType.ANIMATION:
+							anim = InputMediaAnimation(
+								media=j.local_media_path,
+								thumb=j.media_info["thumb"],
+								caption=self.build_signature_caption(job, j.canonical_name),
+								width=j.media_info["width"],
+								height=j.media_info["height"],
+								duration=round(j.media_info["duration"])
+							)
+							tg_chunk.append(anim)
 					mediafs.append(tg_chunk)
 				args["media"] = mediafs
+		elif job.media_type == JobType.TEXT:
+			args["text"] = f"<b>Post text:</b><pre>{job.message_text}</pre>\n\n{self.build_signature_caption(job)}"
+			args["parse_mode"] = ParseMode.HTML
 
 		args["chat_id"] = job.chat_id
 
@@ -362,7 +416,7 @@ class Bot(object):
 			if render_donates:
 				keyboard_buttons[0].append(InlineKeyboardButton("❤ Donate", url=os.environ.get("DONATE_LINK", "https://pay.cryptocloud.plus/pos/W5BMtNQt5bJFoW2E")))
 
-			if keyboard_buttons[0]:  #job.short_text or render_donates:
+			if keyboard_buttons[0]:
 				args["reply_markup"] = InlineKeyboardMarkup(keyboard_buttons)
 
 		return args
@@ -375,10 +429,17 @@ class Bot(object):
 			while not retry_amount >= max_retries:
 				try:
 					reply_message = None
-					if job.media_type in (JobType.VIDEO, JobType.IMAGE, JobType.AUDIO, JobType.ANIMATION):
+					if job.media_type in (JobType.VIDEO, JobType.IMAGE, JobType.AUDIO, JobType.ANIMATION, JobType.TEXT):
+						#if job.media_type in (JobType.VIDEO, JobType.AUDIO):
+						#	await Utils.ensure_me_loaded(self.client)
+						if job.media_type == JobType.TEXT:
+							if job.placeholder_message_id:
+								await self.placeholder.remove(job.chat_id, job.placeholder_message_id)
+								job.placeholder_message_id = None
+
 						if job.placeholder_message_id:
 							try:
-								reply_message = await self.client.edit_message_media(**self.build_tg_args(job))
+								reply_message = await self.editor.edit(**self.build_tg_args(job))
 							except MessageIdInvalid:
 								logging.warning("Placeholder message not found. Looks like placeholder message was deleted by administrator.")
 								job.placeholder_message_id = None
@@ -388,10 +449,17 @@ class Bot(object):
 								JobType.VIDEO: self.client.send_video,
 								JobType.IMAGE: self.client.send_photo,
 								JobType.AUDIO: self.client.send_audio,
-								JobType.ANIMATION: self.client.send_animation
+								JobType.ANIMATION: self.client.send_animation,
+								JobType.TEXT: self.client.send_message
 							}
 							try:
-								reply_message = await send_funcs[job.media_type](**self.build_tg_args(job))
+								while True:
+									try:
+										reply_message = await send_funcs[job.media_type](**self.build_tg_args(job))
+										break
+									except FloodWait as e:
+										logging.warning("FloodWait occurred, waiting '%d' seconds before retry", int(e.value))
+										asyncio.sleep(e.value)
 							except ValueError as e:
 								err_text = str(e)
 								if "Expected" in err_text:
@@ -402,9 +470,10 @@ class Bot(object):
 									job_args[reality.value.lower()] = job_args.pop(expectation.value.lower())
 									reply_message = await send_funcs[reality](**job_args)
 
-						tg_file_id = Utils.extract_file_id(reply_message)
-						tg_file_ids.append(tg_file_id)
-						job.tg_file_id = tg_file_id
+						if reply_message:
+							tg_file_id = Utils.extract_file_id(reply_message)
+							tg_file_ids.append(tg_file_id)
+							job.tg_file_id = tg_file_id
 						logging.info("Uploaded media file with type '%s' tg_file_id is '%s'", job.media_type.value, job.tg_file_id)
 					elif job.media_type == JobType.COLLECTION:
 						col_job_args = self.build_tg_args(job)
@@ -412,7 +481,14 @@ class Bot(object):
 						snd_grp_options = {"chat_id": job.chat_id, "reply_to_message_id": job.message_id}
 						for i, media_chunk in enumerate(col_job_args["media"]):
 							snd_grp_options["media"] = media_chunk
-							messages = await self.client.send_media_group(**snd_grp_options)
+							messages = []
+							while True:
+								try:
+									messages = await self.client.send_media_group(**snd_grp_options)
+									break
+								except FloodWait as e:
+									logging.warning("FloodWait occurred, waiting '%d' seconds before retry", int(e.value))
+									asyncio.sleep(e.value)
 							sent_messages += messages
 							if job.media_collection:
 								for j, _ in enumerate(media_chunk):
